@@ -12,24 +12,38 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+from __future__ import print_function
+
 import copy
 import errno
 import getpass
 import glob
 import hashlib
+import json
 import logging
 import os
 import platform
 import socket
 import subprocess
+import sys
+import tempfile
+import time
 import uuid
 
+from keystoneclient import exceptions as ks_exceptions
+from keystoneclient import auth
+from keystoneclient import session
+from keystoneclient import discover
+from mistralclient.api import client as mistralclient
 from novaclient import client as novaclient
 from novaclient import exceptions
 from oslo_config import cfg
 import psutil
+import pystache
 import six
 
+from instack_undercloud import validator
+
 
 # Making these values properties on a class allows us to delay their lookup,
 # which makes testing code that interacts with these files much easier.
@@ -52,8 +66,6 @@ class Paths(object):
     @property
     def LOG_FILE(self):
         return os.path.expanduser('~/.instack/install-undercloud.log')
-
-
 PATHS = Paths()
 DEFAULT_LOG_LEVEL = logging.DEBUG
 DEFAULT_LOG_FORMAT = '%(asctime)s %(levelname)s: %(message)s'
@@ -73,11 +85,30 @@ secured.
 
 #############################################################################
 """
+CIDR_DEPRECATION_MESSAGE = """
+*****************************************************************************
+The old default CIDR of 192.0.2.0/24 is deprecated due to it being an
+unroutable address range under RFC 5737.  This default will change in the
+Ocata release of OpenStack, so you should stop using the default CIDR and set
+a valid, routable CIDR instead.
+
+Note that if you have already deployed an overcloud with the 192.0.2.0/24
+CIDR, it will not be possible to change it without re-deploying.  If the
+overcloud cannot be re-deployed, you must explicitly set the network values
+in undercloud.conf to ensure continued use of the 192.0.2.0/24 CIDR during
+future upgrades.
+*****************************************************************************
+"""
 # We need 4 GB, leave a little room for variation in what 4 GB means on
 # different platforms.
 REQUIRED_MB = 3750
 
 
+# Allow logging of a warning at the end of the deploy if the deprecated cidr
+# is in use.
+deprecated_cidr = False
+
+
 # When adding new options to the lists below, make sure to regenerate the
 # sample config by running "tox -e genconfig" in the project root.
 _opts = [
@@ -113,14 +144,12 @@ _opts = [
     cfg.StrOpt('undercloud_public_vip',
                default='192.0.2.2',
                help=('Virtual IP address to use for the public endpoints of '
-                     'Undercloud services.  Only used if '
-                     'undercloud_service_certficate is set.')
+                     'Undercloud services. Only used with SSL.')
                ),
     cfg.StrOpt('undercloud_admin_vip',
                default='192.0.2.3',
                help=('Virtual IP address to use for the admin endpoints of '
-                     'Undercloud services.  Only used if '
-                     'undercloud_service_certficate is set.')
+                     'Undercloud services. Only used with SSL.')
                ),
     cfg.StrOpt('undercloud_service_certificate',
                default='',
@@ -128,17 +157,60 @@ _opts = [
                      'connections.  Setting this enables SSL for the '
                      'OpenStack API endpoints, leaving it unset disables SSL.')
                ),
+    cfg.BoolOpt('generate_service_certificate',
+                default=False,
+                help=('When set to True, an SSL certificate will be generated '
+                      'as part of the undercloud install and this certificate '
+                      'will be used in place of the value for '
+                      'undercloud_service_certificate.  The resulting '
+                      'certificate will be written to '
+                      '/etc/pki/tls/certs/undercloud-[undercloud_public_vip].'
+                      'pem.  This certificate is signed by CA selected by the '
+                      '"certificate_generation_ca" option.')
+                ),
+    cfg.StrOpt('certificate_generation_ca',
+               default='local',
+               help=('The certmonger nickname of the CA from which the '
+                     'certificate will be requested. This is used only if '
+                     'the generate_service_certificate option is set. '
+                     'Note that if the "local" CA is selected the '
+                     'certmonger\'s local CA certificate will be extracted to '
+                     '/etc/pki/ca-trust/source/anchors/cm-local-ca.pem and '
+                     'subsequently added to the trust chain.')
+
+               ),
+    cfg.StrOpt('service_principal',
+               default='',
+               help=('The kerberos principal for the service that will use '
+                     'the certificate. This is only needed if your CA '
+                     'requires a kerberos principal. e.g. with FreeIPA.')
+               ),
     cfg.StrOpt('local_interface',
                default='eth1',
                help=('Network interface on the Undercloud that will be '
                      'handling the PXE boots and DHCP for Overcloud '
                      'instances.')
                ),
+    cfg.IntOpt('local_mtu',
+               default=1500,
+               help=('MTU to use for the local_interface.')
+               ),
     cfg.StrOpt('network_cidr',
-               default='192.0.2.0/24',
+               sample_default='192.0.2.0/24',
                help=('Network CIDR for the Neutron-managed network for '
                      'Overcloud instances. This should be the subnet used '
-                     'for PXE booting.')
+                     'for PXE booting. The current default for this value '
+                     'is 192.0.2.0/24, but this is deprecated due to it being '
+                     'a non-routable CIDR under RFC 5737. The default value '
+                     'for this option will be changed in the Ocata release. '
+                     'A different, valid CIDR should be selected to avoid '
+                     'problems. If an overcloud has already been deployed '
+                     'with the 192.0.2.0/24 CIDR and therefore the CIDR '
+                     'cannot be changed, you must set this option to '
+                     '192.0.2.0/24 explicitly to avoid it changing in future '
+                     'releases, and all other network options related to the '
+                     'CIDR (e.g. local_ip) must also be set to maintain a '
+                     'valid configuration.')
                ),
     cfg.StrOpt('masquerade_network',
                default='192.0.2.0/24',
@@ -156,6 +228,24 @@ _opts = [
                help=('End of DHCP allocation range for PXE and DHCP of '
                      'Overcloud instances.')
                ),
+    cfg.StrOpt('hieradata_override',
+               default='',
+               help=('Path to hieradata override file. If set, the file will '
+                     'be copied under /etc/puppet/hieradata and set as the '
+                     'first file in the hiera hierarchy. This can be used to '
+                     'to custom configure services beyond what '
+                     'undercloud.conf provides')
+               ),
+    cfg.StrOpt('net_config_override',
+               default='',
+               help=('Path to network config override template. If set, this '
+                     'template will be used to configure the networking via '
+                     'os-net-config. Must be in json format. '
+                     'Templated tags can be used within the '
+                     'template, see '
+                     'instack-undercloud/elements/undercloud-stack-config/'
+                     'net-config.json.template for example tags')
+               ),
     cfg.StrOpt('inspection_interface',
                default='br-ctlplane',
                deprecated_name='discovery_interface',
@@ -183,7 +273,7 @@ _opts = [
                       'Requires inspection_extras set to True.')
                 ),
     cfg.BoolOpt('inspection_enable_uefi',
-                default=False,
+                default=True,
                 help=('Whether to support introspection of nodes that have '
                       'UEFI-only firmware.')
                 ),
@@ -197,22 +287,31 @@ _opts = [
                 help=('Whether to install Tempest in the Undercloud.')
                 ),
     cfg.BoolOpt('enable_mistral',
-                default=False,
+                default=True,
                 help=('Whether to install Mistral services in the Undercloud.')
                 ),
     cfg.BoolOpt('enable_zaqar',
-                default=False,
+                default=True,
                 help=('Whether to install Zaqar services in the Undercloud.')
                 ),
+    cfg.BoolOpt('enable_telemetry',
+                default=False,
+                help=('Whether to install Telemetry services '
+                      '(ceilometer, aodh) in the Undercloud.')
+                ),
+    cfg.BoolOpt('enable_ui',
+                default=True,
+                help=('Whether to install the TripleO UI.')
+                ),
+    cfg.BoolOpt('enable_validations',
+                default=True,
+                help=('Whether to install requirements to run the TripleO '
+                      'validations.')
+                ),
     cfg.BoolOpt('ipxe_deploy',
                 default=True,
                 help=('Whether to use iPXE for deploy by default.')
                 ),
-    cfg.BoolOpt('enable_monitoring',
-                default=False,
-                help=('Whether to install Monitoring services in the '
-                      'Undercloud.')
-                ),
     cfg.BoolOpt('store_events',
                 default=False,
                 help=('Whether to store events in the Undercloud Ceilometer.')
@@ -224,6 +323,10 @@ _opts = [
                      'greater or equal to the number of bare metal nodes '
                      'you expect to deploy at once to work around '
                      'potential race condition when scheduling.')),
+    cfg.BoolOpt('clean_nodes',
+                default=False,
+                help=('Whether to clean overcloud nodes (wipe the hard drive) '
+                      'between deployments and after the introspection.')),
 ]
 
 # Passwords, tokens, hashes
@@ -316,14 +419,18 @@ _auth_opts = [
                help=('Swift hash suffix. '
                      'If left unset, one will be automatically generated.')
                ),
-    cfg.StrOpt('undercloud_sensu_password',
-               help=('Sensu service password. '
-                     'If left unset, one will be automatically generated.')
-               ),
     cfg.StrOpt('undercloud_haproxy_stats_password',
                help=('HAProxy stats password. '
                      'If left unset, one will be automatically generated.')
                ),
+    cfg.StrOpt('undercloud_zaqar_password',
+               help=('Zaqar password. '
+                     'If left unset, one will be automatically generated.')
+               ),
+    cfg.StrOpt('undercloud_horizon_secret_key',
+               help=('Horizon secret key. '
+                     'If left unset, one will be automatically generated.')
+               ),
 ]
 CONF.register_opts(_opts)
 CONF.register_opts(_auth_opts, group='auth')
@@ -468,6 +575,70 @@ def _check_memory():
         raise RuntimeError('Insufficient memory available')
 
 
+def _check_sysctl():
+    """Check sysctl option availability
+
+    The undercloud will not install properly if some of the expected sysctl
+    values are not available to be set.
+    """
+    options = ['net.ipv4.ip_forward', 'net.ipv4.ip_nonlocal_bind',
+               'net.ipv6.ip_nonlocal_bind']
+
+    not_available = []
+    for option in options:
+        path = '/proc/sys/{opt}'.format(opt=option.replace('.', '/'))
+        if not os.path.isfile(path):
+            not_available.append(option)
+
+    if not_available:
+        LOG.error('Required sysctl options are not available. Check '
+                  'that your kernel is up to date. Missing: '
+                  '{options}'.format(options=", ".join(not_available)))
+        raise RuntimeError('Missing sysctl options')
+
+
+def _validate_network():
+    def error_handler(message):
+        LOG.error('Undercloud configuration validation failed: %s', message)
+        raise validator.FailedValidation(message)
+
+    _validate_cidr()
+    params = {opt.name: CONF[opt.name] for opt in _opts}
+    validator.validate_config(params, error_handler)
+
+
+def _validate_cidr():
+    """Check for default network_cidr
+
+    The old default cidr of 192.0.2.0/24 is deprecated due to being unroutable
+    under RFC 5737.  However, since we need to give users notice of the
+    change, we need some logic to warn them of the problem before actually
+    changing the default, which could be a breaking change on upgrades.
+
+    This function handles the sentinel value of None, which indicates that
+    the user has not overridden the default value, and sets an override on the
+    conf opt to match the previous default.  It also sets a global flag so we
+    can warn about the deprecation at the end of the deploy.
+    """
+    if CONF.network_cidr is None:
+        global deprecated_cidr
+        deprecated_cidr = True
+        CONF.set_override('network_cidr', '192.0.2.0/24')
+
+
+def _validate_configuration():
+    try:
+        _check_hostname()
+        _check_memory()
+        _check_sysctl()
+        _validate_network()
+    except RuntimeError as e:
+        LOG.error('ERROR: An error occured during configuration validation, '
+                  'please check your host configuration and try again. '
+                  '{error}'.format(error=e))
+        sys.exit(1)
+
+
 def _generate_password(length=40):
     """Create a random password
 
@@ -478,148 +649,115 @@ def _generate_password(length=40):
     return hashlib.sha1(uuid_str).hexdigest()[:length]
 
 
+def _get_service_endpoints(name, format_str, public, internal, admin=None,
+                           public_proto='http', internal_proto='http'):
+    endpoints = {}
+    upper_name = name.upper().replace('-', '_')
+    public_port_key = 'port'
+
+    if not admin:
+        admin = internal
+    if public_proto in ['https', 'wss']:
+        public_port_key = 'ssl_port'
+
+    endpoints['UNDERCLOUD_ENDPOINT_%s_PUBLIC' % upper_name] = (
+        format_str % (public_proto, public['host'], public[public_port_key]))
+    endpoints['UNDERCLOUD_ENDPOINT_%s_INTERNAL' % upper_name] = (
+        format_str % (internal_proto, internal['host'], internal['port']))
+    endpoints['UNDERCLOUD_ENDPOINT_%s_ADMIN' % upper_name] = (
+        format_str % (internal_proto, admin['host'], admin['port']))
+    return endpoints
+
+
 def _generate_endpoints(instack_env):
     local_host = instack_env['LOCAL_IP']
     public_host = local_host
-    proto = 'http'
-    heat_public_port = 8004
-    neutron_public_port = 9696
-    glance_public_port = 9292
-    nova_public_port = 8774
-    ceilo_public_port = 8777
-    keystone_public_port = 5000
-    swift_public_port = 8080
-    ironic_public_port = 6385
-    ironic_inspector_public_port = 5050
-    aodh_public_port = 8042
+    public_proto = 'http'
+    internal_host = local_host
+    internal_proto = 'http'
+    zaqar_ws_public_proto = 'ws'
+    zaqar_ws_internal_proto = 'ws'
 
-    if CONF.undercloud_service_certificate:
+    if (CONF.undercloud_service_certificate or
+            CONF.generate_service_certificate):
         public_host = CONF.undercloud_public_vip
-        proto = 'https'
-        heat_public_port = 13004
-        neutron_public_port = 13696
-        glance_public_port = 13292
-        nova_public_port = 13774
-        ceilo_public_port = 13777
-        keystone_public_port = 13000
-        swift_public_port = 13808
-        ironic_public_port = 13385
-        ironic_inspector_public_port = 13050
-        aodh_public_port = 13042
-
-    heat_public_params = (proto, public_host, heat_public_port)
-    heat_internal_params = ('http', local_host, 8004)
-    heat_admin_params = heat_internal_params
-    neutron_public_params = (proto, public_host, neutron_public_port)
-    neutron_internal_params = ('http', local_host, 9696)
-    neutron_admin_params = neutron_internal_params
-    glance_public_params = (proto, public_host, glance_public_port)
-    glance_internal_params = ('http', local_host, 9292)
-    glance_admin_params = glance_internal_params
-    nova_public_params = (proto, public_host, nova_public_port)
-    nova_internal_params = ('http', local_host, 8774)
-    nova_admin_params = nova_internal_params
-    ceilo_public_params = (proto, public_host, ceilo_public_port)
-    ceilo_internal_params = ('http', local_host, 8777)
-    ceilo_admin_params = ceilo_internal_params
-    keystone_public_params = (proto, public_host, keystone_public_port)
-    keystone_internal_params = ('http', local_host, 5000)
-    keystone_admin_params = ('http', local_host, 35357)
-    swift_public_params = (proto, public_host, swift_public_port)
-    swift_internal_params = ('http', local_host, 8080)
-    swift_admin_params = swift_internal_params
-    ironic_public_params = (proto, public_host, ironic_public_port)
-    ironic_internal_params = ('http', local_host, 6385)
-    ironic_admin_params = ironic_internal_params
-    ironic_inspector_public_params = (proto, public_host,
-                                      ironic_inspector_public_port)
-    ironic_inspector_internal_params = ('http', local_host, 5050)
-    ironic_inspector_admin_params = ironic_inspector_internal_params
-    aodh_public_params = (proto, public_host, aodh_public_port)
-    aodh_internal_params = ('http', local_host, 8042)
-    aodh_admin_params = aodh_internal_params
+        public_proto = 'https'
+        zaqar_ws_public_proto = 'wss'
 
     endpoints = {}
 
-    def add_endpoint(name, format_str, public, internal, admin):
-        upper_name = name.upper()
-        endpoints['UNDERCLOUD_ENDPOINT_%s_PUBLIC' % upper_name] = (format_str %
-                                                                   public)
-        endpoints['UNDERCLOUD_ENDPOINT_%s_INTERNAL' % upper_name] = (
-            format_str % internal)
-        endpoints['UNDERCLOUD_ENDPOINT_%s_ADMIN' % upper_name] = (format_str %
-                                                                  admin)
+    endpoint_list = [
+        ('heat',
+            '%s://%s:%d/v1/%%(tenant_id)s',
+            {'host': public_host, 'port': 8004, 'ssl_port': 13004},
+            {'host': internal_host, 'port': 8004}),
+        ('neutron',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 9696, 'ssl_port': 13696},
+            {'host': internal_host, 'port': 9696}),
+        ('glance',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 9292, 'ssl_port': 13292},
+            {'host': internal_host, 'port': 9292}),
+        ('nova',
+            '%s://%s:%d/v2.1',
+            {'host': public_host, 'port': 8774, 'ssl_port': 13774},
+            {'host': internal_host, 'port': 8774}),
+        ('ceilometer',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 8777, 'ssl_port': 13777},
+            {'host': internal_host, 'port': 8777}),
+        ('keystone',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 5000, 'ssl_port': 13000},
+            {'host': internal_host, 'port': 5000},
+            {'host': internal_host, 'port': 35357}),
+        ('swift',
+            '%s://%s:%d/v1/AUTH_%%(tenant_id)s',
+            {'host': public_host, 'port': 8080, 'ssl_port': 13808},
+            {'host': internal_host, 'port': 8080}),
+        ('ironic',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 6385, 'ssl_port': 13385},
+            {'host': internal_host, 'port': 6385}),
+        ('ironic_inspector',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 5050, 'ssl_port': 13050},
+            {'host': internal_host, 'port': 5050}),
+        ('aodh',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 8042, 'ssl_port': 13042},
+            {'host': internal_host, 'port': 8042}),
+        ('mistral',
+            '%s://%s:%d/v2',
+            {'host': public_host, 'port': 8989, 'ssl_port': 13989},
+            {'host': internal_host, 'port': 8989}),
+        ('zaqar',
+            '%s://%s:%d',
+            {'host': public_host, 'port': 8888, 'ssl_port': 13888},
+            {'host': internal_host, 'port': 8888}),
+    ]
+    for endpoint_data in endpoint_list:
+        endpoints.update(
+            _get_service_endpoints(*endpoint_data,
+                                   public_proto=public_proto,
+                                   internal_proto=internal_proto))
+
+    # Zaqar's websocket endpoint
+    # NOTE(jaosorior): Zaqar's websocket endpoint doesn't support being proxied
+    # on a different port. If that's done it will ignore the handshake and
+    # won't work.
+    endpoints.update(_get_service_endpoints(
+        'zaqar-websocket',
+        '%s://%s:%d',
+        {'host': public_host, 'port': 9000, 'ssl_port': 9000},
+        {'host': internal_host, 'port': 9000},
+        public_proto=zaqar_ws_public_proto,
+        internal_proto=zaqar_ws_internal_proto))
 
-    add_endpoint('heat',
-                 '%s://%s:%d/v1/%%(tenant_id)s',
-                 heat_public_params,
-                 heat_internal_params,
-                 heat_admin_params,
-                 )
-    add_endpoint('neutron',
-                 '%s://%s:%d',
-                 neutron_public_params,
-                 neutron_internal_params,
-                 neutron_admin_params,
-                 )
-    add_endpoint('glance',
-                 '%s://%s:%d',
-                 glance_public_params,
-                 glance_internal_params,
-                 glance_admin_params,
-                 )
-    add_endpoint('nova',
-                 '%s://%s:%d/v2/%%(tenant_id)s',
-                 nova_public_params,
-                 nova_internal_params,
-                 nova_admin_params,
-                 )
-    add_endpoint('novav3',
-                 '%s://%s:%d/v3',
-                 nova_public_params,
-                 nova_internal_params,
-                 nova_admin_params,
-                 )
-    add_endpoint('ceilometer',
-                 '%s://%s:%d',
-                 ceilo_public_params,
-                 ceilo_internal_params,
-                 ceilo_admin_params,
-                 )
-    add_endpoint('keystone',
-                 '%s://%s:%d',
-                 keystone_public_params,
-                 keystone_internal_params,
-                 keystone_admin_params,
-                 )
-    add_endpoint('swift',
-                 '%s://%s:%d/v1/AUTH_%%(tenant_id)s',
-                 swift_public_params,
-                 swift_internal_params,
-                 swift_admin_params,
-                 )
     # The swift admin endpoint has a different format from the others
-    endpoints['UNDERCLOUD_ENDPOINT_SWIFT_ADMIN'] = ('%s://%s:%s' %
-                                                    swift_admin_params)
-    add_endpoint('ironic',
-                 '%s://%s:%d',
-                 ironic_public_params,
-                 ironic_internal_params,
-                 ironic_admin_params,
-                 )
-    add_endpoint('ironic_inspector',
-                 '%s://%s:%d',
-                 ironic_inspector_public_params,
-                 ironic_inspector_internal_params,
-                 ironic_inspector_admin_params,
-                 )
-    add_endpoint('aodh',
-                 '%s://%s:%d',
-                 aodh_public_params,
-                 aodh_internal_params,
-                 aodh_admin_params,
-                 )
-
+    endpoints['UNDERCLOUD_ENDPOINT_SWIFT_ADMIN'] = (
+        '%s://%s:%s' % (internal_proto, internal_host, 8080))
     instack_env.update(endpoints)
 
 
@@ -640,6 +778,41 @@ def _write_password_file(instack_env):
             password_file.write('%s=%s\n' % (opt.name, value))
 
 
+def _member_role_exists(instack_env):
+    # This is a workaround for puppet removing the deprecated _member_
+    # role on upgrade - if it exists we must not remove role assignments
+    # or trusts stored in the undercloud heat will break
+    if not _stackrc_exists():
+        instack_env['MEMBER_ROLE_EXISTS'] = 'False'
+        return
+    user, password, tenant, auth_url = _get_auth_values()
+    role_exists = False
+    try:
+        # Note this is made somewhat verbose due to trying to handle
+        # any format auth_url (versionless, v2,0/v3 suffix)
+        auth_plugin_class = auth.get_plugin_class('password')
+        auth_kwargs = {
+            'auth_url': auth_url,
+            'username': user,
+            'password': password,
+            'project_name': tenant}
+        if 'v2.0' not in auth_url:
+            auth_kwargs.update({
+                'project_domain_name': 'default',
+                'user_domain_name': 'default'})
+        auth_plugin = auth_plugin_class(**auth_kwargs)
+        sess = session.Session(auth=auth_plugin)
+        disc = discover.Discover(session=sess)
+        c = disc.create_client()
+        role_names = [r.name for r in c.roles.list()]
+        role_exists = '_member_' in role_names
+    except ks_exceptions.ConnectionError:
+        # This will happen on initial deployment, assume False
+        # as no new deployments should have _member_
+        role_exists = False
+    instack_env['MEMBER_ROLE_EXISTS'] = six.text_type(role_exists)
+
+
 def _generate_environment(instack_root):
     """Generate an environment dict for instack
 
@@ -714,6 +887,9 @@ def _generate_environment(instack_root):
         inspection_kernel_args.append('ipa-debug=1')
     if CONF.inspection_runbench:
         inspection_kernel_args.append('ipa-inspection-benchmarks=cpu,mem,disk')
+    if CONF.inspection_extras:
+        inspection_kernel_args.append('ipa-inspection-dhcp-all-interfaces=1')
+        inspection_kernel_args.append('ipa-collect-lldp=1')
 
     instack_env['INSPECTION_KERNEL_ARGS'] = ' '.join(inspection_kernel_args)
 
@@ -727,13 +903,80 @@ def _generate_environment(instack_root):
     if instack_env.get('DIB_YUM_REPO_CONF'):
         del instack_env['DIB_YUM_REPO_CONF']
 
+    instack_env['TRIPLEO_INSTALL_USER'] = getpass.getuser()
+    instack_env['TRIPLEO_UNDERCLOUD_CONF_FILE'] = PATHS.CONF_PATH
+    instack_env['TRIPLEO_UNDERCLOUD_PASSWORD_FILE'] = PATHS.PASSWORD_PATH
+
     _generate_endpoints(instack_env)
 
     _write_password_file(instack_env)
 
+    if CONF.generate_service_certificate:
+        public_vip = CONF.undercloud_public_vip
+        instack_env['UNDERCLOUD_SERVICE_CERTIFICATE'] = (
+            '/etc/pki/tls/certs/undercloud-%s.pem' % public_vip)
+
+    _member_role_exists(instack_env)
+
     return instack_env
 
 
+def _get_template_path(template):
+    local_template_path = os.path.join(
+        os.path.dirname(__file__),
+        '..',
+        'templates',
+        template)
+    installed_template_path = os.path.join(
+        '/usr/share/instack-undercloud/templates',
+        template)
+    if os.path.exists(local_template_path):
+        return local_template_path
+    else:
+        return installed_template_path
+
+
+def _generate_init_data(instack_env):
+    context = instack_env.copy()
+
+    if CONF.hieradata_override:
+        hiera_entry = os.path.splitext(
+            os.path.basename(CONF.hieradata_override))[0]
+        dst = os.path.join('/etc/puppet/hieradata',
+                           os.path.basename(CONF.hieradata_override))
+        _run_command(['sudo', 'mkdir', '-p', '/etc/puppet/hieradata'])
+        _run_command(['sudo', 'cp', CONF.hieradata_override, dst])
+        _run_command(['sudo', 'chmod', '0644', dst])
+    else:
+        hiera_entry = ''
+
+    if CONF.net_config_override:
+        net_config_json = open(CONF.net_config_override).read()
+    else:
+        net_config_json = \
+            open(_get_template_path('net-config.json.template')).read()
+
+    context['HIERADATA_OVERRIDE'] = hiera_entry
+
+    partials = {'net_config': net_config_json}
+    renderer = pystache.Renderer(partials=partials)
+    template = _get_template_path('config.json.template')
+
+    with open(template) as f:
+        config_json = renderer.render(f.read(), context)
+
+    cfn_path = '/var/lib/heat-cfntools/cfn-init-data'
+    tmp_json = tempfile.mkstemp()[1]
+    with open(tmp_json, 'w') as f:
+        print(config_json, file=f)
+
+    if not os.path.exists(os.path.dirname(cfn_path)):
+        _run_command(['sudo', 'mkdir', '-p', os.path.dirname(cfn_path)])
+
+    _run_command(['sudo', 'mv', tmp_json, cfn_path])
+    _run_command(['sudo', 'chmod', '0644', cfn_path])
+
+
 def _run_instack(instack_env):
     args = ['sudo', '-E', 'instack', '-p', instack_env['ELEMENTS_PATH'],
             '-j', instack_env['JSONFILE'],
@@ -743,13 +986,6 @@ def _run_instack(instack_env):
     LOG.info('Instack completed successfully')
 
 
-def _run_yum_update(instack_env):
-    args = ['sudo', 'yum', 'update', '-y']
-    LOG.info('Running yum update')
-    _run_live_command(args, instack_env, 'yum-update')
-    LOG.info('yum-update completed successfully')
-
-
 def _run_orc(instack_env):
     args = ['sudo', 'os-refresh-config']
     LOG.info('Running os-refresh-config')
@@ -785,7 +1021,7 @@ def _get_auth_values():
     """
     user = _extract_from_stackrc('OS_USERNAME')
     password = _run_command(['sudo', 'hiera', 'admin_password']).rstrip()
-    tenant = _extract_from_stackrc('OS_TENANT')
+    tenant = _extract_from_stackrc('OS_TENANT_NAME')
     auth_url = _extract_from_stackrc('OS_AUTH_URL')
     return user, password, tenant, auth_url
 
@@ -806,6 +1042,17 @@ def _configure_ssh_keys(nova):
             nova.keypairs.create('default', pubkey.read().rstrip())
 
 
+def _delete_default_flavors(nova):
+    """Delete the default flavors from Nova
+
+    The m1.tiny, m1.small, etc. flavors are not useful on an undercloud.
+    """
+    to_delete = ['m1.tiny', 'm1.small', 'm1.medium', 'm1.large', 'm1.xlarge']
+    for f in nova.flavors.list():
+        if f.name in to_delete:
+            nova.flavors.delete(f.id)
+
+
 def _ensure_flavor(nova, name, profile=None):
     try:
         flavor = nova.flavors.create(name, 4096, 1, 40)
@@ -820,9 +1067,20 @@ def _ensure_flavor(nova, name, profile=None):
     LOG.info(message, name, profile)
 
 
+def _stackrc_exists():
+    user_stackrc = os.path.expanduser('~/stackrc')
+    # We gotta check if the copying of stackrc has already been done.
+    if os.path.isfile('/root/stackrc') and os.path.isfile(user_stackrc):
+        return True
+    return False
+
+
 def _copy_stackrc():
     args = ['sudo', 'cp', '/root/stackrc', os.path.expanduser('~')]
-    _run_command(args, name='Copy stackrc')
+    try:
+        _run_command(args, name='Copy stackrc')
+    except subprocess.CalledProcessError:
+        LOG.info("/root/stackrc not found, this is OK on initial deploy")
     args = ['sudo', 'chown', getpass.getuser() + ':',
             os.path.expanduser('~/stackrc')]
     _run_command(args, name='Chown stackrc')
@@ -834,12 +1092,71 @@ def _clean_os_refresh_config():
     _run_command(args, name='Clean os-refresh-config')
 
 
-def _post_config():
+def _create_default_plan(mistral, timeout=180):
+    plan_name = 'overcloud'
+    queue_name = str(uuid.uuid4())
+
+    if plan_name in [env.name for env in mistral.environments.list()]:
+        LOG.info('Not creating default plan "%s" because it already exists.',
+                 plan_name)
+        return
+
+    execution = mistral.executions.create(
+        'tripleo.plan_management.v1.create_default_deployment_plan',
+        workflow_input={'container': plan_name, 'queue_name': queue_name}
+    )
+
+    timeout_at = time.time() + timeout
+
+    while time.time() < timeout_at:
+        exe = mistral.executions.get(execution.id)
+        if exe.state == "RUNNING":
+            time.sleep(5)
+            continue
+        if exe.state == "SUCCESS":
+            return
+        else:
+            raise RuntimeError(
+                "Failed to create the default Deployment Plan. Please check "
+                "the create_default_deployment_plan execution in Mistral with "
+                "`openstack workflow execution list`.")
+    else:
+        exe = mistral.executions.get(execution.id)
+        LOG.error("Timed out waiting for execution %s to finish. State: %s",
+                  exe.id, exe.state)
+        raise RuntimeError(
+            "Timed out creating the default Deployment Plan. Please check "
+            "the create_default_deployment_plan execution in Mistral with "
+            "`openstack workflow execution list`.")
+
+
+def _prepare_ssh_environment(mistral):
+    mistral.executions.create('tripleo.validations.v1.copy_ssh_key')
+
+
+def _post_config_mistral(mistral):
+    _create_default_plan(mistral)
+
+    # Store the snmpd password in a Mistral environment so it can be accessed
+    # by the Mistral actions.
+    snmpd_password = CONF.auth["undercloud_ceilometer_snmpd_password"]
+    mistral.environments.create(
+        name="tripleo.undercloud-config",
+        variables=json.dumps({
+            "undercloud_ceilometer_snmpd_password": snmpd_password
+        }))
+
+    if CONF.enable_validations:
+        _prepare_ssh_environment(mistral)
+
+
+def _post_config(instack_env):
     _copy_stackrc()
     user, password, tenant, auth_url = _get_auth_values()
     nova = novaclient.Client(2, user, password, tenant, auth_url)
 
     _configure_ssh_keys(nova)
+    _delete_default_flavors(nova)
 
     _ensure_flavor(nova, 'baremetal')
     _ensure_flavor(nova, 'control', 'control')
@@ -848,6 +1165,16 @@ def _post_config():
     _ensure_flavor(nova, 'block-storage', 'block-storage')
     _ensure_flavor(nova, 'swift-storage', 'swift-storage')
 
+    if CONF.enable_mistral:
+        mistral_url = instack_env['UNDERCLOUD_ENDPOINT_MISTRAL_PUBLIC']
+        mistral = mistralclient.client(
+            mistral_url=mistral_url,
+            username=user,
+            api_key=password,
+            project_name=tenant,
+            auth_url=auth_url)
+        _post_config_mistral(mistral)
+
 
 def install(instack_root):
     """Install the undercloud
@@ -858,14 +1185,15 @@ def install(instack_root):
     _configure_logging(DEFAULT_LOG_LEVEL, PATHS.LOG_FILE)
     LOG.info('Logging to %s', PATHS.LOG_FILE)
     _load_config()
-    _check_hostname()
-    _check_memory()
     _clean_os_refresh_config()
+    _validate_configuration()
     instack_env = _generate_environment(instack_root)
-    _run_yum_update(instack_env)
+    _generate_init_data(instack_env)
     _run_instack(instack_env)
     _run_orc(instack_env)
-    _post_config()
+    _post_config(instack_env)
     _run_command(['sudo', 'rm', '-f', '/tmp/svc-map-services'], None, 'rm')
     LOG.info(COMPLETION_MESSAGE, {'password_path': PATHS.PASSWORD_PATH,
              'stackrc_path': os.path.expanduser('~/stackrc')})
+    if deprecated_cidr:
+        LOG.warning(CIDR_DEPRECATION_MESSAGE)
